from __future__ import annotations
from pydantic import BaseModel, StrictStr, StrictInt
from typing import Optional

class Message(BaseModel):
    host: Optional[StrictStr]
    port: Optional[StrictInt]
    error: Optional[StrictStr]

    def to_data(self):
        return {
            name: value for name, value in self.__dict__.items() if value is not None
        }



import json
from pydantic import (
    BaseModel,
    Json
)
from typing import (
    Dict, 
    Union, 
    List, 
    TypeVar, 
    Generic,
    Optional
)


T = TypeVar('T', bound=BaseModel)


class Request(Generic[T]):

    def __init__(
        self,
        path: str,
        method: str,
        query: str,
        raw: List[bytes],
        model: Optional[BaseModel] = None
    ) -> None:
        
        self.path = path
        self.method = method
        self._query = query

        self._headers: Dict[str, str] = {}
        self._params: Dict[str, str] = {}
        self._data: Union[str, Json, None] = None

        self._raw = raw
        self._data_line_idx = -1
        self._model = model


    @property
    def headers(self):

        if self._data_line_idx == -1:
            header_lines = self._raw[1:]
            data_line_idx = 0

            for header_line in header_lines:

                if header_line == b'':
                    data_line_idx += 1
                    break
                
                key, value = header_line.decode().split(
                    ':', 
                    maxsplit=1
                )

                self._headers[key.lower()] = value.strip()

                data_line_idx += 1
            
            self._data_line_idx = data_line_idx + 1

        return self._headers
    
    @property
    def params(self) -> Dict[str, str]:
        
        if len(self._params) < 1:
            params = self._query.split('&')

            for param in params:
                key, value = param.split('=')

                self._params[key] = value

        return self._params

    @property
    def body(self):
        
        headers = self.headers

        if self._data is None:
            self._data = b''.join(self._raw[self._data_line_idx:]).strip()

            if headers.get('content-type') == 'application/json':
                self._data = json.loads(self._data)

        return self._data
    
    def data(self) -> Union[bytes, str, Dict[str, str], T]:
        data = self.body

        if isinstance(data, dict) and self._model:
            return self._model(**data)
        
        return data




import json
from enum import Enum
from pydantic import AnyHttpUrl
from typing import Dict, Optional, List, Union
from urllib.parse import urlparse
from .http_message import HTTPMessage
from .message import Message


class HTTPRequestMethod(Enum):
    GET='GET'
    POST='POST'


class HTTPRequest(Message):
    url: AnyHttpUrl
    method: HTTPRequestMethod
    params: Optional[Dict[str, str]]
    headers: Dict[str, str]={}
    data: Optional[Union[str, Message]]

    class Config:
        arbitrary_types_allowed=True

    def prepare_request(self):
        parsed = urlparse(self.url)

        path = parsed.path
        if path is None:
            path = "/"


        if self.params:
            
            params_string = '&'.join([
                f'{name}={value}' for name, value in self.params
            ])

            path = f'{path}?{params_string}'

        request: List[str] = [
            f'{self.method.value} {path} HTTP/1.1'
        ]

        request.append(
            f'host: {parsed.hostname}'
        )

        request.extend([
            f'{key}: {value}' for key, value in self.headers.items()
        ])

        encoded_data = None
        if isinstance(self.data, Message):
            encoded_data = json.dumps(self.data.to_data())

            request.append(
                'content-type: application/msync'
            )

        elif self.data:
            encoded_data = self.data
            content_length = len(encoded_data)
            
            request.append(
                f'content-length: {content_length}'
            )

        request.append('\r\n')

        if encoded_data:
            request.append(encoded_data)

        encoded_request = '\r\n'.join(request)


        return encoded_request.encode()
    
    @classmethod
    def parse(cls, data: bytes):
        response = data.split(b'\r\n')
        
        response_line = response[0]

        headers: Dict[bytes, bytes] = {}

        header_lines = response[1:]
        data_line_idx = 0

        for header_line in header_lines:

            if header_line == b'':
                data_line_idx += 1
                break
            
            key, value = header_line.decode().split(
                ':', 
                maxsplit=1
            )
            headers[key.lower()] = value.strip()

            data_line_idx += 1

        data = b''.join(response[data_line_idx + 1:]).strip()
        

        request_type, status, message = response_line.decode().split(' ')

        return HTTPMessage(
            protocol=request_type,
            status=int(status),
            status_message=message,
            headers=headers,
            data=data.decode()
        )
    
    @classmethod
    def parse_request(cls, data: bytes):
        response = data.split(b'\r\n')
        
        response_line = response[0]

        headers: Dict[bytes, bytes] = {}

        header_lines = response[1:]
        data_line_idx = 0

        for header_line in header_lines:

            if header_line == b'':
                data_line_idx += 1
                break
            
            key, value = header_line.decode().split(
                ':', 
                maxsplit=1
            )
            headers[key.lower()] = value.strip()

            data_line_idx += 1

        data = b''.join(response[data_line_idx + 1:]).strip()
        
        method, path, request_type = response_line.decode().split(' ')

        if path is None or path == '':
            path = "/"

        return HTTPMessage(
            method=method,
            path=path,
            protocol=request_type,
            headers=headers,
            data=data.decode()
        )



from pydantic import (
    BaseModel,
    StrictStr,
    StrictInt
)
from typing import (
    Dict, 
    Union,
    Callable,
    Literal
)


PrimaryType = Union[str, int, float, bytes, bool]


class RegistrarEnv(BaseModel):
    MERCURY_SYNC_REGISTRAR_CLIENT_POLL_RATE: StrictStr='1s'
    MERCURY_SYNC_REGISTRAR_EXPECTED_NODES: StrictInt
    MERCURY_SYNC_REGISTRATION_TIMEOUT: StrictStr='1m'
    MERCURY_SYNC_RESOLVER_CONNECTION_TYPE: Literal["udp", "tcp", "http"]="udp"
    MERCURY_SYNC_RESOLVER_REQUEST_TIMEOUT: StrictStr='5s'
    MERCURY_SYNC_RESOLVER_MAXIMUM_TRIES: StrictInt=5

    @classmethod
    def types_map(self) -> Dict[str, Callable[[str], PrimaryType]]:
        return {
            'MERCURY_SYNC_REGISTRAR_CLIENT_POLL_RATE': str,
            'MERCURY_SYNC_REGISTRAR_EXPECTED_NODES': int,
            'MERCURY_SYNC_REGISTRATION_TIMEOUT': str,
            'MERCURY_SYNC_RESOLVER_CONNECTION_TYPE': str,
            'MERCURY_SYNC_RESOLVER_REQUEST_TIMEOUT': str,
            'MERCURY_SYNC_RESOLVER_MAXIMUM_TRIES': int
        }


import asyncio
import math
import random
import time
from collections import defaultdict, deque
from mercury_sync.env import (
    Env, 
    MonitorEnv,
    load_env
)
from mercury_sync.env.time_parser import TimeParser
from mercury_sync.hooks.client_hook import client
from mercury_sync.hooks.server_hook import server
from mercury_sync.models.healthcheck import HealthCheck, HealthStatus
from mercury_sync.service.controller import Controller
from mercury_sync.snowflake import Snowflake
from mercury_sync.types import Call
from typing import Optional, Dict, Tuple, List, Deque, Union


async def cancel(pending_item: asyncio.Task) -> None:
    pending_item.cancel()
    if not pending_item.cancelled():
        try:
            await pending_item

        except asyncio.CancelledError:
            pass

        except asyncio.IncompleteReadError:
            pass


class Monitor(Controller):

    def __init__(
        self,
        host: str,
        port: int,
        env: Env,
        cert_path: Optional[str]=None,
        key_path: Optional[str]=None,
        workers: int=0,
    ) -> None:
        
        if workers <= 1:
            engine = 'async'

        else:
            engine = 'process'

        if env is None:
            env: Env = load_env(Env)
            
        monitor_env: MonitorEnv = load_env(MonitorEnv)

        super().__init__(
            host,
            port,
            cert_path=cert_path,
            key_path=key_path,
            workers=workers,
            env=env,
            engine=engine
        )

        self.status: HealthStatus = 'initializing'
        

        self.error_context: Optional[str] = None

        self.registration_timeout = TimeParser(
            monitor_env.MERCURY_SYNC_REGISTRATION_TIMEOUT
        ).time

        self.boot_wait = TimeParser(
            monitor_env.MERCURY_SYNC_BOOT_WAIT
        ).time
        
        self._healthchecks: Dict[str, asyncio.Task] = {}
        self._registered: Dict[int, Tuple[str, int]] = {}
        self._running = False

        self._cleanup_interval = TimeParser(
            env.MERCURY_SYNC_CLEANUP_INTERVAL
        ).time

        self._poll_interval = TimeParser(
            monitor_env.MERCURY_SYNC_HEALTH_POLL_INTERVAL
        ).time

        self._poll_timeout = TimeParser(
            monitor_env.MERCURY_SYNC_HEALTH_CHECK_TIMEOUT
        ).time

        self._reboot_timeout = TimeParser(
            monitor_env.MERCURY_SYNC_IDLE_REBOOT_TIMEOUT
        ).time

        self._max_time_idle = TimeParser(
            monitor_env.MERCURY_SYNC_MAX_TIME_IDLE
        ).time

        self._poll_retries = monitor_env.MERCURY_SYNC_MAX_POLL_MULTIPLIER

        self._sync_interval = TimeParser(
            monitor_env.MERCURY_SYNC_UDP_SYNC_INTERVAL
        ).time

        self._check_nodes_count = monitor_env.MERCURY_SYNC_INDIRECT_CHECK_NODES

        self.min_suspect_multiplier = monitor_env.MERCURY_SYNC_MIN_SUSPECT_TIMEOUT_MULTIPLIER
        self.max_suspect_multiplier = monitor_env.MERCURY_SYNC_MAX_SUSPECT_TIMEOUT_MULTIPLIER 
        self._min_suspect_node_count = monitor_env.MERCURY_SYNC_MIN_SUSPECT_NODES_THRESHOLD
        self._max_poll_multiplier = monitor_env.MERCURY_SYNC_MAX_POLL_MULTIPLIER
        self._local_health_multiplier = 0

        self._confirmed_suspicions: Dict[Tuple[str, int], int] = defaultdict(lambda: 0)
        self._waiter: Union[asyncio.Future, None] = None

        self._tasks_queue: Deque[asyncio.Task] = deque()
        self._degraded_nodes: Deque[Tuple[str, int]] = deque()
        self._suspect_nodes: Deque[Tuple[str, int]] = deque()

        self._degraded_tasks: Dict[Tuple[str, int], asyncio.Task] = {}
        self._suspect_tasks: Dict[Tuple[str, int], asyncio.Task] = {}
        self._latest_update: Dict[Tuple[str,int], int] = {}

        self._local_health_monitor: Union[asyncio.Task, None] = None
        self._udp_sync_task: Union[asyncio.Task, None] = None
        self._tcp_sync_task: Union[asyncio.Task, None] = None

        self._cleanup_task: Union[asyncio.Task, None] = None
        self._investigating_nodes: Dict[Tuple[str, int], Dict[Tuple[str, int]]] = defaultdict(dict)
        self._node_statuses: Dict[Tuple[str, int], HealthStatus] = {}

        self.bootstrap_host: Union[str, None] = None
        self.bootstrap_port: Union[int, None] = None

        self._healthy_statuses = [
            'waiting',
            'healthy'
        ]

        self._unhealthy_statuses = [
            'suspect',
            'failed'
        ]

    @server()
    async def update_node_status(
        self,
        shard_id: int,
        healthcheck: HealthCheck
    ) -> Call[HealthCheck]:
        
        update_node_host = healthcheck.source_host
        update_node_port = healthcheck.source_port
        update_status = healthcheck.status

        if healthcheck.target_host and healthcheck.target_port:
            update_node_host = healthcheck.target_host
            update_node_port = healthcheck.target_port

        if healthcheck.target_status:
            update_status = healthcheck.target_status
        
        local_status = self._node_statuses.get((
            update_node_host,
            update_node_port
        ))


        target_last_updated: Union[int, None] = healthcheck.target_last_updated
        local_last_updated: Union[int, None] = self._latest_update.get((
            update_node_host,
            update_node_port
        ), 0)

        not_self = update_node_host != self.host and update_node_port != self.port

        if not_self and local_status in self._unhealthy_statuses:

            self._tasks_queue.append(
                asyncio.create_task(
                    self.refresh_clients(
                        HealthCheck(
                            host=update_node_host,
                            port=update_node_port,
                            source_host=self.host,
                            source_port=self.port,
                            status=update_status,
                            error=healthcheck.error
                        )
                    )
                )
            )

        elif not_self and local_status is None:

            self._tasks_queue.append(
                asyncio.create_task(
                    self.extend_client(
                        HealthCheck(
                            host=update_node_host,
                            port=update_node_port,
                            source_host=self.host,
                            source_port=self.port,
                            status=self.status,
                            error=healthcheck.error
                        )
                    )
                )
            )

        if target_last_updated > local_last_updated:
            self._node_statuses[(update_node_host, update_node_port)] = update_status

        return HealthCheck(
            host=healthcheck.source_host,
            port=healthcheck.source_port,
            source_host=self.host,
            source_port=self.port,
            status=self.status
        )

    @server()
    async def update_as_suspect(
        self,
        shard_id: int,
        healthcheck: HealthCheck
    ) -> Call[HealthCheck]:
        
        source_host = healthcheck.source_host
        source_port = healthcheck.source_port

        if self.status == 'healthy':
            self._local_health_multiplier = min(
                self._local_health_multiplier, 
                self._max_poll_multiplier
            ) + 1

            self._tasks_queue.append(
                asyncio.create_task(
                    self._run_healthcheck(
                        source_host,
                        source_port
                    )
                )
            )

        return HealthCheck(
            host=source_host,
            port=source_port,
            source_host=self.host,
            source_port=self.port,
            status=self.status
        )

    @server()
    async def send_indirect_check(
        self,
        shard_id: int,
        healthcheck: HealthCheck
    ) -> Call[HealthCheck]:
        
        source_host = healthcheck.source_host
        source_port = healthcheck.source_port

        target_host = healthcheck.target_host
        target_port = healthcheck.target_port

        try:

            investigation_update = self._acknowledge_indirect_probe(
                source_host,
                source_port,
                target_host,
                target_port
            )

            indirect_probe = self._run_healthcheck(
                target_host,
                target_port
            )

            for task in asyncio.as_completed([
                investigation_update,
                indirect_probe
            ]):
                await task

                self._local_health_multiplier = max(
                    self._local_health_multiplier - 1, 
                    0
                )

            # We've received a refutation
            return HealthCheck(
                host=healthcheck.source_host,
                port=healthcheck.source_port,
                target_status=self._node_statuses.get((target_host, target_port)),   
                source_host=target_host,
                source_port=target_port,
                status=self.status,
                error=self.error_context
            )

        except asyncio.TimeoutError:
            self._local_health_multiplier = min(
                self._local_health_multiplier, 
                self._max_poll_multiplier
            ) + 1

            # Our suspicion is correct!
            return HealthCheck(
                host=healthcheck.source_host,
                port=healthcheck.source_port,
                source_host=target_host,
                source_port=target_port,
                target_status='suspect', 
                status=self.status
            )
    
    @server()
    async def update_acknowledged(
        self,
        shard_id: int,
        healthcheck: HealthCheck
    ) -> Call[HealthCheck]:
        source_host = healthcheck.source_host
        source_port = healthcheck.source_port
        target_host = healthcheck.target_host
        target_port = healthcheck.target_port

        if self._investigating_nodes.get((target_host, target_port)) is None:
            self._investigating_nodes[(target_host, target_port)] = {}

        self._investigating_nodes[(target_host, target_port)].update({
            (source_host, source_port): healthcheck.status
        })

        return HealthCheck(
            host=source_host,
            port=source_port,
            source_host=self.host,
            source_port=self.port,
            status=self.status
        )
        
    @server()
    async def register_health_update(
        self,
        shard_id: int,
        healthcheck: HealthCheck
    ) -> Call[HealthCheck]:
            
        source_host = healthcheck.source_host
        source_port = healthcheck.source_port

        target_host = healthcheck.target_host
        target_port = healthcheck.target_port

        local_node_status = self._node_statuses.get((source_host, source_port))

        suspect_tasks = dict(self._suspect_tasks)
        suspect_task = suspect_tasks.pop((source_host, source_port), None)

        if suspect_task:
            self._tasks_queue.append(
                asyncio.create_task(
                    cancel(suspect_task)
                )
            )

            del self._suspect_tasks[(source_host, source_port)]

            self._suspect_tasks = suspect_tasks


        if local_node_status is None:

            self._tasks_queue.append(
                asyncio.create_task(
                    self.extend_client(
                        HealthCheck(
                            host=source_host,
                            port=source_port,
                            source_host=self.host,
                            source_port=self.port,
                            status=healthcheck.status,
                            error=healthcheck.error
                        )
                    )
                )
            )

        elif local_node_status in self._unhealthy_statuses:
            self._tasks_queue.append(
                asyncio.create_task(
                    self.refresh_clients(
                        HealthCheck(
                            host=source_host,
                            port=source_port,
                            source_host=self.host,
                            source_port=self.port,
                            status=healthcheck.status,
                            error=healthcheck.error
                        )
                    )
                )
            )
    

        self._node_statuses[(source_host, source_port)] = healthcheck.status
        self._latest_update[(source_host, source_port)] = Snowflake.parse(shard_id).timestamp

        if target_host and target_port: 
            self._node_statuses[(target_host, target_port)] = healthcheck.target_status

        return HealthCheck(
            host=healthcheck.source_host,
            port=healthcheck.source_port,
            source_host=self.host,
            source_port=self.port,
            source_status=local_node_status,
            error=self.error_context,
            status=self.status
        )

    @client('register_health_update')
    async def push_health_update(
        self,
        host: str,
        port: int,
        health_status: HealthStatus,
        target_host: Optional[str]=None,
        target_port: Optional[str]=None,
        error_context: Optional[str]=None
    ) -> Call[HealthCheck]:
        
        target_status: Union[HealthCheck, None] = None
        if target_host and target_port:
            target_status = self._node_statuses.get((target_host, target_port))

        return HealthCheck(
            host=host,
            port=port,
            source_host=self.host,
            source_port=self.port,
            target_host=target_host,
            target_port=target_port,
            target_status=target_status,
            error=error_context,
            status=health_status
        )
    
    @client('register_health_update', as_tcp=True)
    async def push_tcp_health_update(
        self,
        host: str,
        port: int,
        health_status: HealthStatus,
        target_host: Optional[str]=None,
        target_port: Optional[str]=None,
        error_context: Optional[str]=None
    ) -> Call[HealthCheck]:
        
        target_status: Union[HealthCheck, None] = None
        if target_host and target_port:
            target_status = self._node_statuses.get((target_host, target_port))

        return HealthCheck(
            host=host,
            port=port,
            source_host=self.host,
            source_port=self.port,
            target_host=target_host,
            target_port=target_port,
            target_status=target_status,
            error=error_context,
            status=health_status
        )
    
    async def _run_tcp_healthcheck(
        self, 
        host: str, 
        port: int,
        target_host: Optional[str]=None,
        target_port: Optional[str]=None
    ) -> Union[Tuple[int, HealthCheck], None]:
        
        shard_id: Union[int, None] = None
        healthcheck: Union[HealthCheck, None] = None

        for _ in range(self._poll_retries):

            try:

                response: Tuple[int, HealthCheck] = await asyncio.wait_for(
                    self.push_tcp_health_update(
                        host,
                        port,
                        self.status,
                        target_host=target_host,
                        target_port=target_port,
                        error_context=self.error_context
                    ),
                    timeout=self._poll_timeout * (self._local_health_multiplier + 1)
                )

                shard_id, healthcheck = response
                source_host, source_port = healthcheck.source_host, healthcheck.source_port

                self._node_statuses[(source_host, source_port)] = healthcheck.status

                self._local_health_multiplier = max(
                    0, 
                    self._local_health_multiplier - 1
                )

                return shard_id, healthcheck

            except (asyncio.TimeoutError, RuntimeError):
                self._local_health_multiplier = min(
                    self._local_health_multiplier, 
                    self._max_poll_multiplier
                ) + 1
                
        check_host = host
        check_port = port

        if target_host and target_port:
            check_host = target_host
            check_port = target_port

        if healthcheck is None and self._node_statuses.get((check_host, check_port)) == 'healthy':
            self._node_statuses[(check_host, check_port)] = 'suspect'

            self._suspect_nodes.append((
                check_host,
                check_port
            ))

            self._suspect_tasks[(host, port)] = asyncio.create_task(
                self._start_suspect_monitor()
            )

        return shard_id, healthcheck
    
    @client('update_acknowledged')
    async def push_acknowledge_check(
        self,
        host: str,
        port: int,
        target_host: str,
        target_port: int,
        health_status: HealthStatus,
        error_context: Optional[str]=None
    ) -> Call[HealthCheck]:
        return HealthCheck(
            host=host,
            port=port,
            source_host=self.host,
            source_port=self.port,
            target_host=target_host,
            target_port=target_port,
            status=health_status,
            error=error_context
        )
    
    @client('send_indirect_check')
    async def request_indirect_check(
        self,
        host: str,
        port: int,
        target_host: str,
        target_port: int,
        health_status: HealthStatus,
        error_context: Optional[str]=None

    ) -> Call[HealthCheck]:
        return HealthCheck(
            host=host,
            port=port,
            target_host=target_host,
            target_port=target_port,
            target_status=self._node_statuses[(target_host, target_port)],
            source_host=self.host,
            source_port=self.port,
            error=error_context,
            status=health_status
        )
    
    @client('update_node_status')
    async def push_status_update(
        self,
        host: str,
        port: int,
        health_status: HealthStatus,
        target_host: Optional[str]=None,
        target_port: Optional[int]=None,
        error_context: Optional[str]=None
    ) -> Call[HealthCheck]:
        
        target_status: Union[HealthStatus, None] = None
        target_last_updated: Union[int,  None] = self._latest_update.get(
            (host, port), 0
        )

        if target_host and target_port:
            target_status = self._node_statuses.get((target_host, target_port))
            target_last_updated = self._latest_update.get(
                (target_host, target_port), 0
            )

        return HealthCheck(
            host=host,
            port=port,
            source_host=self.host,
            source_port=self.port,
            target_host=target_host,
            target_port=target_port,
            target_last_updated=target_last_updated,
            target_status=target_status,
            status=health_status,
            error=error_context
        )
    
    @client('update_node_status', as_tcp=True)
    async def push_tcp_status_update(
        self,
        host: str,
        port: int,
        health_status: HealthStatus,
        target_host: Optional[str]=None,
        target_port: Optional[int]=None,
        error_context: Optional[str]=None
    ) -> Call[HealthCheck]:
        
        target_status: Union[HealthStatus, None] = None
        target_last_updated: Union[int,  None] = self._latest_update.get(
            (host, port), 0
        )

        if target_host and target_port:
            target_status = self._node_statuses.get((target_host, target_port))
            target_last_updated = self._latest_update.get(
                (target_host, target_port), 0
            )


        return HealthCheck(
            host=host,
            port=port,
            source_host=self.host,
            source_port=self.port,
            target_host=target_host,
            target_port=target_port,
            target_status=target_status,
            target_last_updated=target_last_updated,
            status=health_status,
            error=error_context
        )
    
    @client('update_as_suspect')
    async def push_suspect_update(
        self,
        host: str,
        port: int,
        health_status: HealthStatus,
        error_context: Optional[str]=None
    ) -> Call[HealthCheck]:
        return HealthCheck(
            host=host,
            port=port,
            source_host=self.host,
            source_port=self.port,
            status=health_status,
            error=error_context
        )
    
    async def start(self):

        await self.start_server()
        await asyncio.sleep(self.boot_wait)

    async def register(
        self,
        host: str,
        port: int
    ):
        
        self.bootstrap_host = host
        self.bootstrap_port = port
        
        await asyncio.wait_for(
            asyncio.create_task(
                self.start_client(
                    HealthCheck(
                        host=host,
                        port=port,
                        source_host=self.host,
                        source_port=self.port,
                        status=self.status
                    ),
                    cert_path=self.cert_path,
                    key_path=self.key_path
                )
            ),
            timeout=self.registration_timeout
        )

        self._node_statuses[(host, port)] = 'healthy'

        self.status = 'healthy'
        self._running = True
        
        self._healthchecks[(host, port)] = asyncio.create_task(
            self.start_health_monitor()
        )
        
        self.confirmation_task = asyncio.create_task(
            self.cleanup_pending_checks()
        )

        self._udp_sync_task = asyncio.create_task(
            self._run_udp_state_sync()
        )

        self._tcp_sync_task = asyncio.create_task(
            self._run_tcp_state_sync()
        )

    def _calculate_min_suspect_timeout(self):
        nodes_count = len(self._node_statuses) + 1

        poll_interval = self._poll_interval * (self._local_health_multiplier + 1)

        return round(
            self.min_suspect_multiplier * math.log10(nodes_count) * poll_interval,
            2
        )

    def _calculate_max_suspect_timeout(self, min_suspect_timeout: float):
        
        return round(
            self.max_suspect_multiplier * min_suspect_timeout,
            2
        )

    def _calculate_suspicion_timeout(
        self,
        suspect_node_address: Tuple[str, int]
    ):

        min_suspect_timeout = self._calculate_min_suspect_timeout()
        
        max_suspect_timeout = self._calculate_max_suspect_timeout(min_suspect_timeout)

        confirmed_suspect_count = max(
            0,
            self._confirmed_suspicions[suspect_node_address]
        )

        timeout_modifier = math.log(
            confirmed_suspect_count + 1
        )/math.log(self._min_suspect_node_count + 1)

        timeout_difference = max_suspect_timeout - min_suspect_timeout

        return max(
            min_suspect_timeout,
            max_suspect_timeout - (timeout_difference * timeout_modifier)
        )
    
    async def _acknowledge_indirect_probe(
        self,
        host: str,
        port: int,
        target_host: str,
        target_port: int
    ):
        try:
            timeout = self._poll_timeout * (self._local_health_multiplier + 1)
            response: Tuple[int, HealthCheck] = await asyncio.wait_for(
                self.push_acknowledge_check(
                    host,
                    port,
                    target_host,
                    target_port,
                    self.status,
                    error_context=self.error_context
                ),
                timeout=timeout
            )

            _, healthcheck = response

            source_host, source_port = healthcheck.source_host, healthcheck.source_port

            self._node_statuses[(source_host, source_port)] = healthcheck.status

            self._local_health_multiplier = max(
                0, 
                self._local_health_multiplier - 1
            )

        except asyncio.TimeoutError:
            self._local_health_multiplier = min(
                self._local_health_multiplier, 
                self._max_poll_multiplier
            ) + 1

            if self._node_statuses.get((host, port)) == 'healthy':
                self._node_statuses[(host, port)] = 'suspect'

                self._suspect_nodes.append((
                    host,
                    port
                ))

                self._suspect_tasks[(host, port)] = asyncio.create_task(
                    self._start_suspect_monitor()
                )

    async def _run_healthcheck(
        self, 
        host: str, 
        port: int,
        target_host: Optional[str]=None,
        target_port: Optional[str]=None
    ) -> Union[Tuple[int, HealthCheck], None]:
        
        shard_id: Union[int, None] = None
        healthcheck: Union[HealthCheck, None] = None

        for _ in range(self._poll_retries):

            try:

                response: Tuple[int, HealthCheck] = await asyncio.wait_for(
                    self.push_health_update(
                        host,
                        port,
                        self.status,
                        target_host=target_host,
                        target_port=target_port,
                        error_context=self.error_context
                    ),
                    timeout=self._poll_timeout * (self._local_health_multiplier + 1)
                )

                shard_id, healthcheck = response
                source_host, source_port = healthcheck.source_host, healthcheck.source_port

                self._node_statuses[(source_host, source_port)] = healthcheck.status

                self._local_health_multiplier = max(
                    0, 
                    self._local_health_multiplier - 1
                )

                return shard_id, healthcheck

            except asyncio.TimeoutError:
                self._local_health_multiplier = min(
                    self._local_health_multiplier, 
                    self._max_poll_multiplier
                ) + 1

        check_host = host
        check_port = port

        if target_host and target_port:
            check_host = target_host
            check_port = target_port

        if healthcheck is None and self._node_statuses.get((check_host, check_port)) == 'healthy':
            self._node_statuses[(check_host, check_port)] = 'suspect'

            self._suspect_nodes.append((
                check_host,
                check_port
            ))

            self._suspect_tasks[(host, port)] = asyncio.create_task(
                self._start_suspect_monitor()
            )

        return shard_id, healthcheck

    async def _start_suspect_monitor(self):

        if len(self._suspect_nodes) < 1:
            return
        
        address = self._suspect_nodes.pop()
        suspect_host, suspect_port = address
        
        suspicion_timeout = self._calculate_suspicion_timeout(address)


        elapsed = 0
        start = time.monotonic()

        while elapsed < suspicion_timeout and self._node_statuses[(suspect_host, suspect_port)] == 'suspect':

            self._tasks_queue.append(
                asyncio.create_task(
                    self._push_suspect_update(
                        host=suspect_host,
                        port=suspect_port,
                        health_status=self.status,
                        error_context=self.error_context
                    )
                )
            )
                
            confirmation_members = self._get_confirmation_members((
                suspect_host,
                suspect_port
            ))

            suspect_count = await self._request_indirect_probe(
                suspect_host,
                suspect_port,
                confirmation_members
            )

            self._confirmed_suspicions[(suspect_host, suspect_port)] += max(
                0,
                suspect_count - 1
            )

            indirect_ack_count = len(self._investigating_nodes[(suspect_host, suspect_port)])

            missing_ack_count = len(confirmation_members) - indirect_ack_count

            next_health_multiplier = self._local_health_multiplier + missing_ack_count - indirect_ack_count
            if next_health_multiplier < 0:
                self._local_health_multiplier = 0

            else:
                self._local_health_multiplier = min(
                    next_health_multiplier, 
                    self._max_poll_multiplier
                ) + 1

            confirmation_members_count = len(confirmation_members)

            if suspect_count < confirmation_members_count:
                # We had a majority confirmation the node was healthy.
                self._investigating_nodes[(suspect_host, suspect_port)] = {}
                self._confirmed_suspicions[(suspect_host, suspect_port)] = 0

                self._node_statuses[(suspect_host, suspect_port)] = 'healthy'

                break
            
            await asyncio.sleep(
                self._poll_interval * (self._local_health_multiplier + 1)
            ) 

            elapsed = time.monotonic() - start
            suspicion_timeout = self._calculate_suspicion_timeout(address)

        if self._node_statuses[(suspect_host, suspect_port)] == 'suspect':
            self._node_statuses[(suspect_host, suspect_port)] = 'failed'
        
        self._investigating_nodes[(suspect_host, suspect_port)] = {}
        self._confirmed_suspicions[(suspect_host, suspect_port)] = 0
            
    def _get_confirmation_members(self, suspect_address: Tuple[str, int]) -> List[Tuple[str, int]]:

        confirmation_members = [
            address for address in self._node_statuses.keys() if address != suspect_address
        ]

        confirmation_members_count = len(confirmation_members)

        if self._check_nodes_count > confirmation_members_count:
            self._check_nodes_count = confirmation_members_count
            
        confirmation_members = random.sample(
            confirmation_members, 
            self._check_nodes_count
        )

        return confirmation_members

    async def _request_indirect_probe(
        self,
        host: str,
        port: int,
        confirmation_members: List[Tuple[str, int]]
    ) -> Tuple[List[Call[HealthCheck]], int]:

        if len(confirmation_members) < 1:
            requested_checks = [
                asyncio.create_task(
                    self._run_tcp_healthcheck(
                        host,
                        port
                    )
                )
            ]
        
        else:
            requested_checks = [
                asyncio.create_task(
                    self.request_indirect_check(
                        node_host,
                        node_port,
                        host,
                        port,
                        self.status,
                        error_context=self.error_context
                    )
                ) for node_host, node_port in confirmation_members
            ]

            requested_checks.append(
                asyncio.create_task(
                    self._run_tcp_healthcheck(
                        host,
                        port
                    )
                )
            )

        check_tasks: Tuple[List[asyncio.Task], List[asyncio.Task]] = await asyncio.wait(
            requested_checks, 
            timeout=self._poll_timeout * (self._local_health_multiplier + 1)
        )

        completed, pending = check_tasks

        results: List[Call[HealthCheck]]  = await asyncio.gather(
            *completed,
            return_exceptions=True
        )

        healthchecks = [
            result for result in results if isinstance(
                result,
                tuple
            )
        ]

        errors = [
            result for result in results if isinstance(
                result,
                tuple
            ) is False
        ]

        sorted_checks: List[Call[HealthCheck]] = list(sorted(
            healthchecks,
            key=lambda check: Snowflake.parse(
                Snowflake.parse(check[0]).timestamp
            ).timestamp
        ))

        suspect = [
            (
                shard_id,
                check
            ) for shard_id, check in sorted_checks if check.target_status == 'suspect'
        ]

        healthy = [
            (
                shard_id,
                check
            ) for shard_id, check in sorted_checks if check.target_status == 'healthy'
        ]
            
        if len(healthy) < 1:
            suspect_count = len(suspect) + len(pending) + len(errors)

        else:
            suspect_checks: List[Call[HealthCheck]] = []
            for suspect_shard_id, suspect_check in suspect:

                newer_count = 0
                for healthy_shard_id, _ in healthy:
                    if suspect_shard_id > healthy_shard_id:
                        newer_count += 1

                if newer_count >= len(healthy):
                    suspect_checks.append((
                        suspect_shard_id,
                        suspect_check
                    ))

            suspect_count = len(suspect_checks) + len(pending) + len(errors)
        
        await asyncio.gather(*[
            cancel(pending_check) for pending_check in pending
        ])

        return suspect_count
    
    async def _propagate_state_update(
        self,
        target_host: str,
        target_port: int
    ):
        monitoring = [
            address for address, status in self._node_statuses.items() if status in self._healthy_statuses
        ]

        for host, port in monitoring:
            await self.push_health_update(
                host,
                port,
                self.status,
                target_host=target_host,
                target_port=target_port
            )

    async def run_forever(self):
        self._waiter = asyncio.Future()
        await self._waiter

    async def start_health_monitor(self):

        while self._running:

            monitors = list(self._node_statuses.keys())

            host: Union[str, None] = None
            port: Union[int, None] = None

            monitors_count = len(monitors)

            if monitors_count > 0:
                host, port = random.choice(monitors)

            if self._node_statuses.get((host, port)) == 'healthy':
        
                self._tasks_queue.append(
                    asyncio.create_task(
                        self._run_healthcheck(
                            host,
                            port
                        )
                    )
                )

            await asyncio.sleep(
                self._poll_interval * (self._local_health_multiplier + 1)
            )

    async def _run_udp_state_sync(self):
        while self._running:

            monitors = [
                address for address, status in self._node_statuses.items() if status in self._healthy_statuses
            ]

            active_nodes_count = len(monitors)

            if active_nodes_count > 0:

                self._tasks_queue.extend([
                    asyncio.create_task(
                        self._push_state_to_node(
                            host=host,
                            port=port
                        )
                    ) for host, port in monitors
                ])

            await asyncio.sleep(
                self._sync_interval
            )

    async def _run_tcp_state_sync(self):

        await asyncio.sleep(
            self._sync_interval/2
        )

        while self._running:

            monitors = [
                address for address, status in self._node_statuses.items() if status in self._healthy_statuses
            ]

            active_nodes_count = len(monitors)

            if active_nodes_count > 0:

                self._tasks_queue.extend([
                    asyncio.create_task(
                        self._push_state_to_node_tcp(
                            host=host,
                            port=port
                        )
                    ) for host, port in monitors
                ])

            await asyncio.sleep(
                self._sync_interval
            )

    async def _push_state_to_node(
        self,
        host: str,
        port: int
    ):
        
        updates = [
            self._push_status_update(
                host=host,
                port=port,
                target_host=node_host,
                target_port=node_port
            ) for node_host, node_port in self._node_statuses if self._node_statuses.get((
                node_host,
                node_port
            )) == 'healthy' and host != node_host and port != node_port
        ]

        if len(updates) > 0:
            await asyncio.gather(*updates)

    async def _push_state_to_node_tcp(
        self,
        host: str,
        port: int
    ):
        updates = [
            self._push_tcp_status_update(
                host=host,
                port=port,
                target_host=node_host,
                target_port=node_port
            ) for node_host, node_port in self._node_statuses if self._node_statuses.get((
                node_host,
                node_port
            )) == 'healthy' and host != node_host and port != node_port
        ]
    
        if len(updates) > 0:
            await asyncio.gather(*updates)

    async def _push_status_update(
        self,
        host: str,
        port: int,
        target_host: Optional[str]=None,
        target_port: Optional[int]=None
    ) -> Tuple[
            Union[int, None], 
            Union[HealthCheck, None]
        ]:
        
        shard_id: Union[int, None] = None
        healthcheck: Union[HealthCheck, None] = None

        for _ in range(self._poll_retries):

            try:

                response: Tuple[int, HealthCheck] = await asyncio.wait_for(
                    self.push_status_update(
                        host,
                        port,
                        self.status,
                        target_host=target_host,
                        target_port=target_port,
                        error_context=self.error_context
                    ),
                    timeout=self._poll_timeout * (self._local_health_multiplier + 1)
                )

                shard_id, healthcheck = response
                source_host, source_port = healthcheck.source_host, healthcheck.source_port

                self._node_statuses[(source_host, source_port)] = healthcheck.status

                return shard_id, healthcheck

            except asyncio.TimeoutError:
                pass

        return shard_id, healthcheck

    async def _push_tcp_status_update(
        self,
        host: str,
        port: int,
        target_host: Optional[str]=None,
        target_port: Optional[int]=None
    ):
        shard_id: Union[int, None] = None
        healthcheck: Union[HealthCheck, None] = None
        
        try:
            
            response: Tuple[int, HealthCheck] = await asyncio.wait_for(
                self.push_tcp_status_update(
                    host,
                    port,
                    self.status,
                    target_host=target_host,
                    target_port=target_port,
                    error_context=self.error_context
                ),
                timeout=self._poll_timeout * (self._local_health_multiplier + 1)
            )

            shard_id, healthcheck = response
            source_host, source_port = healthcheck.source_host, healthcheck.source_port

            self._node_statuses[(source_host, source_port)] = healthcheck.status

            return shard_id, healthcheck

        except asyncio.TimeoutError:
            pass

        return shard_id, healthcheck


    async def _push_suspect_update(
        self,
        host: str,
        port: int,
        health_status: HealthStatus,
        error_context: Optional[str]=None
    ):
        
        try:
            response: Tuple[int, HealthCheck] = await asyncio.wait_for(
                self.push_suspect_update(
                    host=host,
                    port=port,
                    health_status=health_status,
                    error_context=error_context
                ),
                timeout=self._poll_timeout * (self._local_health_multiplier + 1)
            )

            _, healthcheck = response

            self._node_statuses[(host, port)] = healthcheck.status

        except asyncio.TimeoutError:
            pass

    async def cleanup_pending_checks(self):

        while self._running:

            for pending_check in list(self._tasks_queue):
                if pending_check.done() or pending_check.cancelled():
                    try:
                        await pending_check

                    except (
                        ConnectionRefusedError,
                        ConnectionAbortedError,
                        ConnectionResetError
                    ):
                        pass

                    self._tasks_queue.remove(pending_check)

            await asyncio.sleep(self._cleanup_interval)
    
    async def shutdown(self):
        self._running = False
        self._local_health_monitor.cancel()

        await asyncio.gather(*[
            cancel(check) for check in self._tasks_queue
        ])

        await asyncio.gather(*[
            cancel(remote_check) for remote_check in self._healthchecks.values()
        ])

        await cancel(self._local_health_monitor)
        
        await cancel(self._cleanup_task)

        await cancel(self._udp_sync_task)

        await cancel(self._tcp_sync_task)

        await self.close()

    async def soft_shutdown(self):
        await asyncio.gather(*[
            cancel(check) for check in self._tasks_queue
        ])


            
            


    





import json
from pydantic import (
    StrictStr,
    StrictInt,
    Json
)
from typing import Dict, Optional, Union, Literal, List
from .message import Message


class HTTPMessage(Message):
    protocol: StrictStr
    path: Optional[StrictStr]
    method: Optional[
        Literal[
            "GET", 
            "POST",
            "HEAD",
            "OPTIONS", 
            "PUT", 
            "PATCH", 
            "DELETE"
        ]
    ]
    status: Optional[StrictInt]
    status_message: Optional[StrictStr]
    params: Dict[StrictStr, StrictStr]={}
    headers: Dict[StrictStr, StrictStr]={}
    data: Optional[Union[Json, StrictStr]]

    def prepare_response(self):

        request: List[str] = [
            f'HTTP/1.1 {self.status} {self.error}'
        ]

        request.extend([
            f'{key}: {value}' for key, value in self.headers.items()
        ])

        encoded_data = None
        if isinstance(self.data, Message):
            encoded_data = json.dumps(self.data.to_data()).encode()

        elif self.data:
            encoded_data = self.data.encode()
            content_length = len(encoded_data)
            
            request.append(
                f'content-length: {content_length}'
            )

        else:
            request.append(
                'content-length: 0'
            )

        request.append('\r\n')
        if encoded_data:
            request.append(encoded_data)


        encoded_request = '\r\n'.join(request)

        return encoded_request.encode()


import asyncio
import socket
from mercury_sync.discovery.dns.core.url import URL
from mercury_sync.env import load_env, Env
from mercury_sync.hooks import (
    client,
    server
)
from mercury_sync.models.dns_message import DNSMessage
from mercury_sync.models.dns_message_group import DNSMessageGroup
from mercury_sync.service.controller import Controller
from mercury_sync.discovery.dns.core.record import (
    Record
)
from mercury_sync.discovery.dns.core.random import RandomIDGenerator

from mercury_sync.discovery.dns.resolver import DNSResolver
from mercury_sync.models.dns_entry import DNSEntry
from mercury_sync.models.service import Service
from mercury_sync.types import Call
from typing import (
    Optional, 
    List,
    Union,
    Dict,
    Tuple
)


class Registrar(Controller):

    def __init__(
        self,
        host: str,
        port: int,
        cert_path: Optional[str]=None,
        key_path: Optional[str]=None,
        workers: int=0,
        env: Env=None
    ) -> None:

        if env is None:
            env = load_env(Env)
    
        super().__init__(
            host,
            port,
            cert_path=cert_path,
            key_path=key_path,
            env=env,
            workers=workers,
            engine='async'
        )

        self.resolver = DNSResolver(
            host,
            port,
            self._instance_id,
            self._env
        )

        self.random_id_generator = RandomIDGenerator()

        self._nameservers: List[URL] = []
        self._next_nameserver_idx = 0
        self._connected_namservers: Dict[Tuple[str, int], bool] = {}
        self._connected_domains: Dict[str, bool] = {}

    def add_entries(
        self,
        entries: List[DNSEntry]
    ):

        for entry in entries:
            for domain, record in entry.to_record_data():
                
                self.resolver.add_to_cache(
                    domain,
                    record.rtype,
                    record
                )

    async def add_nameservers(
        self,
        urls: List[str]
    ):
        urls = self.resolver.add_nameservers(urls)

        await self.resolver.connect_nameservers(
            urls,
            cert_path=self.cert_path,
            key_path=self.key_path
        )

        self._nameservers.extend(urls)

    def _next_nameserver_url(self) -> Union[URL, None]:

        if len(self._nameservers) > 0:

            namserver_url = self._nameservers[self._next_nameserver_idx]

            self._next_nameserver_idx = (
                self._next_nameserver_idx + 1
            )%len(self._nameservers)

            return namserver_url

    @server()
    async def update_registered(
        self,
        shard_id: int,
        registration: DNSMessage
    ):
        for record in registration.query_domains:
            self.resolver.add_to_cache(
                record.name,
                record.record_type,
                record.data
            )

        return registration


    @server()
    async def resolve_query(
        self,
        shard_id: int,
        query: DNSMessage
    ) -> Call[DNSMessage]:
        
        messages: List[DNSMessage] = []
        
        for record in query.query_domains:
                  
            dns_message, has_result = await self.resolver.query(
                record.name,
                record_type=record.record_type
            )

            if has_result is False:
                # TODO: Query using client.
                pass

            dns_data = dns_message.to_data()
            dns_data.update({
                'query_id': query.query_id,
                'has_result': has_result
            })

            response = DNSMessage(**dns_data)

            messages.append(response)

        return DNSMessageGroup(
            messages=messages
        )
            
    @client('resolve_query')
    async def submit_query(
        self,
        host: str,
        port: int,
        entry: DNSEntry
    ) -> Call[DNSMessageGroup]:
   
        return DNSMessage(
            host=host,
            port=port,
            query_domains=[
                Record(
                    name=domain,
                    record_type=record.rtype,
                    data=record,
                    ttl=entry.time_to_live

                ) for domain, record in entry.to_record_data()
            ]
        )
    
    @client('update_registered')
    async def submt_registration(
        self,
        host: str,
        port: int,
        entry: DNSEntry
    ) -> Call[DNSMessage]:
        return DNSMessage(
            host=host,
            port=port,
            query_domains=[
                Record(
                    name=domain,
                    record_type=record.rtype,
                    data=record,
                    ttl=entry.time_to_live

                ) for domain, record in entry.to_record_data()
            ]
        )
    
    async def query(
        self,
        entry: DNSEntry
    ) -> List[DNSEntry]:

        nameserver_url = self._next_nameserver_url()

        host = nameserver_url.host
        port = nameserver_url.port

        if nameserver_url.ip_type is not None:
            host = socket.gethostbyname(nameserver_url.host)

        if not self._connected_namservers.get((host, port)):
            
            await self.start_client(
                DNSMessage(
                    host=host,
                    port=port
                )
            )

            self._connected_namservers[(host, port)] = True

        _, results = await self.submit_query(
            host,
            port,
            entry
        )

        entries: List[DNSEntry]=[]

        for message in results.messages:
            for answer in message.query_answers:

                entries.append(
                    DNSEntry.from_record_data(
                        answer.name,
                        answer.data
                    )
                )
                
        return entries
    
    async def register(
        self,
        entry: DNSEntry
    ) -> List[DNSEntry]:
        
        nameserver_url = self._next_nameserver_url()

        host = nameserver_url.host
        port = nameserver_url.port

        if nameserver_url.ip_type is not None:
            host = socket.gethostbyname(nameserver_url.host)

        if not self._connected_namservers.get((host, port)):
            
            await self.start_client(
                DNSMessage(
                    host=host,
                    port=port
                )
            )

            self._connected_namservers[(host, port)] = True

        _, results = await self.submt_registration(
            host,
            port,
            entry
        )

        entries: List[DNSEntry]=[]

        for answer in results.query_domains:

                entries.append(
                    DNSEntry.from_record_data(
                        answer.name,
                        answer.data
                    )
                )
                
        return entries\
        
    async def discover(self, url: str):

        ptr_records = await self.get_ptr_records(url)

        srv_records = await self.get_srv_records(ptr_records)
        txt_records = await self.get_txt_records(ptr_records)  

        services_data: Dict[str, Dict[str, Union[str, int,  Dict[str, str]]]] = {}
        services: Dict[str, Service] = {}

        for record in srv_records:
            service_url = record.to_domain(record.record_type.name)


            services_data[service_url] = {
                "service_instance": record.instance_name,
                "service_name": record.service_name,
                "service_protocol": record.domain_protocol,
                'service_url': service_url,
                'service_ip': record.domain_targets[0],
                'service_port': record.domain_port,
                'service_context': {}
            }


        for record in txt_records:
            service_url = record.domain_name

            services_data[service_url]['service_context'].update(record.domain_values)

        for service_url, data in services_data.items():
            services[service_url] = Service(**data)


        return list(services.values())

    async def get_ptr_records(
        self,
        url: str
    ):

        (
            service_name, 
            domain_protocol,
            domain_name
        ) = DNSEntry.to_ptr_segments(url)

    
        return await self.query(
            DNSEntry(
                service_name=service_name,
                domain_protocol=domain_protocol,
                domain_name=domain_name,
                record_types=["PTR"]
            )
        )
    
    async def get_srv_records(
        self,
        ptr_records: List[DNSEntry]
    ):
        
        srv_records: List[List[DNSEntry]] = await asyncio.gather(*[
            self.query(
                DNSEntry(
                    instance_name=entry.instance_name,
                    service_name=entry.service_name,
                    domain_protocol=entry.domain_protocol,
                    domain_name=entry.domain_name,
                    record_types=["SRV"]
                )
            ) for entry in ptr_records
        ], return_exceptions=True)

        service_records: List[DNSEntry] = []

        for results in srv_records:
            service_records.extend(results)

        return service_records
    
    async def get_txt_records(
        self,
        ptr_records: List[DNSEntry]
    ):
        txt_records = await asyncio.gather(*[
            self.query(
                DNSEntry(
                    instance_name=entry.instance_name,
                    service_name=entry.service_name,
                    domain_protocol=entry.domain_protocol,
                    domain_name=entry.domain_name,
                    record_types=["TXT"]
                )
            ) for entry in ptr_records
        ], return_exceptions=True)

        text_records: List[DNSEntry] = []
        for results in txt_records:
            text_records.extend(results)

        return text_records



from __future__ import annotations
import socket
import struct
from typing import Dict, Iterable, Tuple
from .record_data import RecordData
from .record_types import RecordType
from .utils import (
    load_domain_name,
    pack_domain_name
)


class SOARecordData(RecordData):

    def __init__(self, *args):
        super().__init__(
            RecordType.SOA,
            data=args
        )

        (
            mname,
            rname,
            serial,
            refresh,
            retry,
            expire,
            minimum,
        ) = args

        self.mname = mname
        self.rname = rname
        self.serial = serial
        self.refresh = refresh
        self.retry = retry
        self.expire = expire
        self.minimum = minimum

    def __repr__(self):
        return '<%s: %s>' % (self.type_name, self.rname)


    @classmethod
    def load(
        cls, 
        data: bytes, 
        cursor_position: int, 
        size: int
    ) -> Tuple[int, SOARecordData]:
        
        cursor_position, mname = load_domain_name(data, cursor_position)
        cursor_position, rname = load_domain_name(data, cursor_position)

        (
            serial,
            refresh,
            retry,
            expire,
            minimum,
        ) = struct.unpack(
            '!LLLLL', 
            data[cursor_position:cursor_position + 20]
        )

        return cursor_position + 20, SOARecordData(
            mname, 
            rname, 
            serial, 
            refresh, 
            retry, 
            expire,
            minimum
        )

    def dump(self, names: Dict[str, int], offset: int) -> Iterable[bytes]:

        mname = pack_domain_name(
            self.mname, 
            names, 
            offset + 2
        )

        mname_length = len(mname)

        domain_name = pack_domain_name(
            self.rname, 
            names, 
            offset + 2 + mname_length
        )

        record_bytes = struct.pack(
            '!LLLLL', 
            self.serial, 
            self.refresh, 
            self.retry,
            self.expire, 
            self.minimum
        )

        record_data = [
            mname,
            domain_name, 
            record_bytes
        ]

        for data in record_data:
            yield data



import struct
from typing import Union


def pack_string(string: Union[str, bytes], btype='B') -> bytes:
    '''Pack string into `{length}{data}` format.'''
    if not isinstance(string, bytes):
        string = string.encode()
    length = len(string)
    return struct.pack('%s%ds' % (btype, length), length, string)



from __future__ import annotations
from typing import Tuple
from .domain_record_data import DomainRecordData
from .record_types import RecordType
from .utils import load_domain_name


class CNAMERecordData(DomainRecordData):

    def __init__(self, data: str):
        super().__init__(
            RecordType.CNAME,
            data=data
        )

    @classmethod
    def load(
        cls, 
        data: bytes, 
        cursor_position: int,
        size: int
    ) -> Tuple[int, CNAMERecordData]:
        
        cursor_position, domain = load_domain_name(
            data,
            cursor_position
        )

        return cursor_position, CNAMERecordData(domain)


import asyncio
from concurrent.futures import ThreadPoolExecutor


class BackupVolume:

    def __init__(
        self,
        path: str,
        service_name: str,
        instance_id: str
    ) -> None:
        self.path = path
        self.service_name = service_name
        self.instance_id = instance_id

    




from ipaddress import IPv4Address
from pydantic import (
    BaseModel,
    StrictStr,
    StrictInt,
    StrictBool,
    IPvAnyAddress
)
from typing import (
    Dict, 
    Union,
    Callable,
    Literal
)


PrimaryType = Union[str, int, float, bytes, bool]


class Env(BaseModel):
    MERCURY_SYNC_HTTP_RATE_LIMIT_STRATEGY: Literal["none", "global", "endpoint", "ip-address"]="none"
    MERCURY_SYNC_HTTP_RATE_LIMIT_PERIOD: StrictStr='1s'
    MERCURY_SYNC_HTTP_POOL_SIZE: StrictInt=10
    MERCURY_SYNC_USE_HTTP_MSYNC_ENCRYPTION: StrictBool=False
    MERCURY_SYNC_USE_HTTP_SERVER: StrictBool=False
    MERCURY_SYNC_USE_HTTP_AND_TCP_SERVERS: StrictBool=False
    MERCURY_SYNC_TCP_CONNECT_RETRIES: StrictInt=3
    MERCURY_SYNC_CLEANUP_INTERVAL: StrictStr='10s'
    MERCURY_SYNC_MAX_CONCURRENCY: StrictInt=2048
    MERCURY_SYNC_AUTH_SECRET: StrictStr
    MERCURY_SYNC_MULTICAST_GROUP: IPvAnyAddress='224.1.1.1'

    @classmethod
    def types_map(self) -> Dict[str, Callable[[str], PrimaryType]]:
        return {
            'MERCURY_SYNC_HTTP_RATE_LIMIT_STRATEGY': str,
            'MERCURY_SYNC_HTTP_RATE_LIMIT_PERIOD': str,
            'MERCURY_SYNC_USE_TCP_SERVER': lambda value: True if value.lower() == 'true' else False,
            'MERCURY_SYNC_HTTP_POOL_SIZE': int,
            'MERCURY_SYNC_USE_HTTP_MSYNC_ENCRYPTION': lambda value: True if value.lower() == 'true' else False,
            'MERCURY_SYNC_USE_HTTP_SERVER': lambda value: True if value.lower() == 'true' else False,
            'MERCURY_SYNC_TCP_CONNECT_RETRIES': int,
            'MERCURY_SYNC_CLEANUP_INTERVAL': str,
            'MERCURY_SYNC_MAX_CONCURRENCY': int,
            'MERCURY_SYNC_AUTH_SECRET': str,
            'MERCURY_SYNC_MULTICAST_GROUP': str
        }


def load_string(
    buffer: bytes, 
    offset: int
):
    '''Load a character string from packed data.'''
    length = buffer[offset]
    offset += 1
    data = buffer[offset:offset + length]
    return offset + length, data



import asyncio
from contextlib import AbstractAsyncContextManager
from mercury_sync.models.request import Request
from mercury_sync.models.http_message import HTTPMessage
from types import TracebackType
from typing import (
    Dict, 
    Optional, 
    Type,
    Callable,
    Coroutine,
    Tuple
)



class LeakyBucketLimiter(AbstractAsyncContextManager):

    __slots__ = (
        "max_rate",
        "time_period",
        "_rate_per_sec",
        "_level",
        "_last_check",
        "_waiters",
    )

    max_rate: float
    time_period: float

    def __init__(self, max_rate: float, time_period: float = 60) -> None:
        self.max_rate = max_rate
        self.time_period = time_period
        self._rate_per_sec = max_rate / time_period
        self._level = 0.0
        self._last_check = 0.0

        self._waiters: Dict[asyncio.Task, asyncio.Future] = {}
        self._loop = asyncio.get_event_loop()

    def _leak(self) -> None:

        loop = asyncio.get_running_loop()

        if self._level:
            elapsed = loop.time() - self._last_check
            decrement = elapsed * self._rate_per_sec
            self._level = max(self._level - decrement, 0)

        self._last_check = loop.time()

    def has_capacity(self, amount: float = 1) -> bool:

        self._leak()
        requested = self._level + amount

        if requested < self.max_rate:

            for fut in self._waiters.values():
                if not fut.done():
                    fut.set_result(True)
                    break

        return self._level + amount <= self.max_rate

    async def acquire(
        self, 
        amount: float = 1
    ) -> None:
        if amount > self.max_rate:
            raise ValueError("Can't acquire more than the maximum capacity")

   
        task = asyncio.current_task(
            loop=self._loop
        )

        assert task is not None

        while not self.has_capacity(amount):

            fut = self._loop.create_future()
            self._waiters[task] = fut

            try:


                await asyncio.wait_for(
                    asyncio.shield(fut), 
                    timeout=(1 / self._rate_per_sec * amount)
                )

            except asyncio.TimeoutError:
                pass

            fut.cancel()

        self._waiters.pop(task, None)
        self._level += amount

        return None

    async def __aenter__(self) -> None:
        await self.acquire()
        return None

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc: Optional[BaseException],
        tb: Optional[TracebackType],
    ) -> None:
        return None


import time
from mercury_sync.discovery.dns.core.url import URL
from typing import List, Iterable, Union
from .exceptions import NoNameServer


class NameServer:

    def __init__(
        self, 
        urls: List[Union[str, URL]]
    ):
        self.data = [
            URL(url) if isinstance(url, str) else url for url in urls
        ]
        

        self._failures = [0] * len(self.data)
        self.timestamp = 0
        self._update()

    def __bool__(self):
        return len(self.data) > 0

    def __iter__(self):
        return iter(self.data)

    def iter(self) -> Iterable[URL]:
        if not self.data: 
            raise NoNameServer()
        
        return iter(self.data)

    def _update(self):

        if time.time() > self.timestamp + 60:
            self.timestamp = time.time()

            self._sorted = list(
                self.data[i] for i in sorted(
                    range(len(self.data)), 
                    key=lambda i: self._failures[i]
                )
            )

            self._failures = [0] * len(self.data)

    def success(self, item):
        self._update()

    def fail(self, item):
        self._update()
        index = self.data.index(item)
        self._failures[index] += 1



import ipaddress
from typing import List


class SubnetRange:

    def __init__(
        self,
        base_address: str,
        subnet_range: int=24
    ) -> None:
        self.subnet = f'{base_address}/{subnet_range}'
        self._network = ipaddress.ip_network(self.subnet, strict=False)
        self._addresses = [
            str(ip) for ip in self._network.hosts()
        ]

        self.reserved: List[str] = []

    def __iter__(self):

        available_addresses = [
            address for address in self._addresses if address not in self.reserved
        ]

        for address in available_addresses:
                yield address


from __future__ import annotations
from typing import Dict, Iterable, Optional, Tuple
from .record_types import (
    RecordType,
    RecordTypesMap
)




class RecordData:
    '''Base class of RData'''

    def __init__(
        self,
        rtype: RecordType,
        data: Optional[str]=None
    ) -> None:
        self.types_map = RecordTypesMap()
        self.rtype = rtype
        self.data = data

    def __hash__(self):
        return hash(self.data)

    def __eq__(self, other: RecordData):
        return self.__class__ == other.__class__ and self.data == other.data

    def __repr__(self):
        return '<%s: %s>' % (self.type_name, self.data)

    @property
    def type_name(self):
        return self.types_map.names_mapping.get(
            self.rtype
        ).lower()

    @classmethod
    def load(cls, data: bytes, ip_length: int, size: int) -> Tuple[int, RecordData]:
        raise NotImplementedError

    def dump(self, names: Dict[str, int], offset: int) -> Iterable[bytes]:
        raise NotImplementedError




import asyncio
import pickle
import socket
import ssl
import zstandard
from collections import deque, defaultdict
from mercury_sync.connection.base.connection_type import ConnectionType
from mercury_sync.encryption import AESGCMFernet
from mercury_sync.env import Env
from mercury_sync.env.time_parser import TimeParser
from mercury_sync.models.message import Message
from mercury_sync.snowflake.snowflake_generator import SnowflakeGenerator
from typing import (
    Tuple, 
    Deque, 
    Any, 
    Dict, 
    Coroutine, 
    AsyncIterable,
    Union,
    Optional
)
from mercury_sync.connection.tcp.protocols import (
    MercurySyncTCPClientProtocol,
    MercurySyncTCPServerProtocol
)


class MercurySyncTCPConnection:

    def __init__(
        self,
        host: str,
        port: int,
        instance_id: int,
        env: Env
    ) -> None:

        self.id_generator = SnowflakeGenerator(instance_id)
        self.env = env

        self.host = host
        self.port = port

        self.events: Dict[
            str,
            Coroutine
        ] = {}

        self.queue: Dict[str, Deque[Tuple[str, int, float, Any]] ] = defaultdict(deque)
        self.parsers: Dict[str, Message] = {}
        self.connected = False
        self._running = False

        self._client_transports: Dict[str, asyncio.Transport] = {}
        self._server: asyncio.Server = None
        self._loop: Union[asyncio.AbstractEventLoop, None] = None
        self._waiters: Dict[str, Deque[asyncio.Future]] = defaultdict(deque)
        self._pending_responses: Deque[asyncio.Task]= deque()
        self._last_call: Deque[str] = deque()

        self._sent_values = deque()
        self.server_socket = None
        self._stream = False
        
        self._client_key_path: Union[str, None] = None
        self._client_cert_path: Union[str, None] = None

        self._server_key_path: Union[str, None] = None
        self._server_cert_path: Union[str, None] = None 
        
        self._client_ssl_context: Union[ssl.SSLContext, None] = None
        self._server_ssl_context: Union[ssl.SSLContext, None] = None
        
        self._encryptor = AESGCMFernet(env)
        self._semaphore: Union[asyncio.Semaphore, None] = None
        self._compressor: Union[zstandard.ZstdCompressor, None] = None
        self._decompressor: Union[zstandard.ZstdDecompressor, None] = None
        self._cleanup_task: Union[asyncio.Task, None] = None
        self._sleep_task: Union[asyncio.Task, None] = None
        self._cleanup_interval = TimeParser(env.MERCURY_SYNC_CLEANUP_INTERVAL).time
        self._max_concurrency = env.MERCURY_SYNC_MAX_CONCURRENCY
        self._tcp_connect_retries = env.MERCURY_SYNC_TCP_CONNECT_RETRIES

        self.connection_type = ConnectionType.TCP

    def connect(
        self,
        cert_path: Optional[str]=None,
        key_path: Optional[str]=None,
        worker_socket: Optional[socket.socket]=None
    ):

        try:

            self._loop = asyncio.get_event_loop()

        except Exception:
            self._loop = asyncio.new_event_loop()
            asyncio.set_event_loop(self._loop)
            
        self._running = True
        self._semaphore = asyncio.Semaphore(self._max_concurrency)

        self._compressor = zstandard.ZstdCompressor()
        self._decompressor = zstandard.ZstdDecompressor()

        if cert_path and key_path:
            self._server_ssl_context = self._create_server_ssl_context(
                cert_path=cert_path,
                key_path=key_path
            ) 

        if self.connected is False and worker_socket is None:
            self.server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            self.server_socket.bind((self.host, self.port))

            self.server_socket.setblocking(False)

        elif self.connected is False:
            self.server_socket = worker_socket
            host, port = worker_socket.getsockname()
            
            self.host = host
            self.port = port

        if self.connected is False:

            server = self._loop.create_server(
                lambda: MercurySyncTCPServerProtocol(
                    self.read
                ),
                sock=self.server_socket,
                ssl=self._server_ssl_context
            )

            self._server = self._loop.run_until_complete(server)

            self.connected = True

            self._cleanup_task = self._loop.create_task(self._cleanup())

    async def connect_async(
        self,
        cert_path: Optional[str]=None,
        key_path: Optional[str]=None,
        worker_socket: Optional[socket.socket]=None
    ):

        try:

            self._loop = asyncio.get_event_loop()

        except Exception:
            self._loop = asyncio.new_event_loop()
            asyncio.set_event_loop(self._loop)

        self._running = True
        self._semaphore = asyncio.Semaphore(self._max_concurrency)

        self._compressor = zstandard.ZstdCompressor()
        self._decompressor = zstandard.ZstdDecompressor()

        if cert_path and key_path:
            self._server_ssl_context = self._create_server_ssl_context(
                cert_path=cert_path,
                key_path=key_path
            ) 

        if self.connected is False and worker_socket is None:
            self.server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

            try:
                self.server_socket.bind((self.host, self.port))

            except Exception:
                pass


            self.server_socket

            self.server_socket.setblocking(False)

        elif self.connected is False:
            self.server_socket = worker_socket
            host, port = worker_socket.getsockname()

            self.host = host
            self.port = port

        if self.connected is False:

            server = self._loop.create_server(
                lambda: MercurySyncTCPServerProtocol(
                    self.read
                ),
                sock=self.server_socket,
                ssl=self._server_ssl_context
            )

            self._server = await server
            self.connected = True

            self._cleanup_task = self._loop.create_task(self._cleanup())

    def _create_server_ssl_context(
        self, 
        cert_path: Optional[str]=None,
        key_path: Optional[str]=None
    ) -> ssl.SSLContext:
        
        if self._server_cert_path is None:
            self._server_cert_path = cert_path

        if self._server_key_path is None:
            self._server_key_path = key_path

        ssl_ctx = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
        ssl_ctx.options |= ssl.OP_NO_TLSv1
        ssl_ctx.options |= ssl.OP_NO_TLSv1_1
        ssl_ctx.options |= ssl.OP_SINGLE_DH_USE
        ssl_ctx.options |= ssl.OP_SINGLE_ECDH_USE
        ssl_ctx.load_cert_chain(cert_path, keyfile=key_path)
        ssl_ctx.load_verify_locations(cafile=cert_path)
        ssl_ctx.check_hostname = False
        ssl_ctx.verify_mode = ssl.VerifyMode.CERT_REQUIRED
        ssl_ctx.set_ciphers('ECDHE-ECDSA-AES256-GCM-SHA384:ECDHE-RSA-AES256-GCM-SHA384')

        return ssl_ctx

    async def connect_client(
        self,
        address: Tuple[str, int],
        cert_path: Optional[str]=None,
        key_path: Optional[str]=None,
        worker_socket: Optional[socket.socket]=None,
    ) -> None:
        
        if self._semaphore is None:
            self._semaphore = asyncio.Semaphore(self._max_concurrency)
        
        self._loop = asyncio.get_event_loop()
        if cert_path and key_path:
            self._client_ssl_context = self._create_client_ssl_context(
                cert_path=cert_path,
                key_path=key_path
            ) 

        if worker_socket is None:

            tcp_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            tcp_socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
            await self._loop.run_in_executor(None, tcp_socket.connect, address)

            tcp_socket.setblocking(False)

        else:

            tcp_socket = worker_socket

        last_error: Union[Exception, None] = None

        for _ in range(self._tcp_connect_retries):

            try:

                client_transport, _ = await self._loop.create_connection(
                    lambda: MercurySyncTCPClientProtocol(
                        self.read
                    ),
                    sock=tcp_socket,
                    ssl=self._client_ssl_context
                )

                self._client_transports[address] = client_transport

                return client_transport
            
            except ConnectionRefusedError as connection_error:
                last_error = connection_error

            await asyncio.sleep(1)

        if last_error:
            raise last_error
    
    def _create_client_ssl_context(
        self, 
        cert_path: Optional[str]=None,
        key_path: Optional[str]=None
    ) -> ssl.SSLContext:
        
        if self._client_cert_path is None:
            self._client_cert_path = cert_path

        if self._client_key_path is None:
            self._client_key_path = key_path

        ssl_ctx = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
        ssl_ctx.options |= ssl.OP_NO_TLSv1
        ssl_ctx.options |= ssl.OP_NO_TLSv1_1
        ssl_ctx.load_cert_chain(cert_path, keyfile=key_path)
        ssl_ctx.load_verify_locations(cafile=cert_path)
        ssl_ctx.check_hostname = False
        ssl_ctx.verify_mode = ssl.VerifyMode.CERT_REQUIRED
        ssl_ctx.set_ciphers('ECDHE-ECDSA-AES256-GCM-SHA384:ECDHE-RSA-AES256-GCM-SHA384')

        return ssl_ctx
    
    async def _cleanup(self):
        while self._running:
            self._sleep_task = asyncio.create_task(
                asyncio.sleep(self._cleanup_interval)
            )

            await self._sleep_task

            for pending in list(self._pending_responses):
                if pending.done() or pending.cancelled():

                    try:
                        await pending

                    except Exception:
                        await self.close()
                        await self.connect_async(
                            cert_path=self._client_cert_path,
                            key_path=self._client_key_path
                        )

                    self._pending_responses.pop()

    async def send(
        self, 
        event_name: bytes,
        data: bytes, 
        address: Tuple[str, int]
    ) -> Tuple[int, Dict[str, Any]]:
        
        async with self._semaphore:
            self._last_call.append(event_name)

            client_transport = self._client_transports.get(address)
            if client_transport is None:
                await self.connect_client(
                    address,
                    cert_path=self._client_cert_path,
                    key_path=self._client_key_path
                )

                client_transport = self._client_transports.get(address)

            item = pickle.dumps(
                (
                    'request',
                    self.id_generator.generate(),
                    event_name,
                    data,
                    self.host,
                    self.port
                ),
                protocol=pickle.HIGHEST_PROTOCOL
            )

            encrypted_message = self._encryptor.encrypt(item)
            compressed = self._compressor.compress(encrypted_message)

            client_transport.write(compressed)

            waiter = self._loop.create_future()
            self._waiters[event_name].append(waiter)

            (
                _,
                shard_id,
                _,
                response_data,
                _, 
                _
            ) = await waiter

            return (
                shard_id,
                response_data
            )
        
    async def send_bytes(
        self,
        event_name: str,
        data: bytes,
        address: Tuple[str, int]
    ) -> bytes:
        async with self._semaphore:
            self._last_call.append(event_name)

            client_transport = self._client_transports.get(address)
            if client_transport is None:
                await self.connect_client(
                    address,
                    cert_path=self._client_cert_path,
                    key_path=self._client_key_path
                )

                client_transport = self._client_transports.get(address)

            client_transport.write(data)

            waiter = self._loop.create_future()
            self._waiters[event_name].append(waiter)
            return await waiter
    
    async def stream(
        self, 
        event_name: str,
        data: Any, 
        address: Tuple[str, int]
    ) -> AsyncIterable[Tuple[int, Dict[str, Any]]]: 
        
        async with self._semaphore:
            self._last_call.append(event_name)

            client_transport = self._client_transports.get(address)


            if self._stream is False:
                item = pickle.dumps(
                    (
                        'stream_connect',
                        self.id_generator.generate(),
                        event_name,
                        data,
                        self.host,
                        self.port
                    ),
                    protocol=pickle.HIGHEST_PROTOCOL
                )


            else:
                item = pickle.dumps(
                    (
                        'stream',
                        self.id_generator.generate(),
                        event_name,
                        data,
                        self.host,
                        self.port
                    ),
                    protocol=pickle.HIGHEST_PROTOCOL
                )

            encrypted_message = self._encryptor.encrypt(item)
            compressed = self._compressor.compress(encrypted_message)

            client_transport.write(compressed)

            waiter = self._loop.create_future()
            self._waiters[event_name].append(waiter)

            await waiter

            if self._stream is False:

                self.queue[event_name].pop()

                self._stream = True

                item = pickle.dumps(
                    (
                        'stream',
                        self.id_generator.generate(),
                        event_name,
                        data,
                        self.host,
                        self.port
                    ),
                    pickle.HIGHEST_PROTOCOL
                )

                encrypted_message = self._encryptor.encrypt(item)
                compressed = self._compressor.compress(encrypted_message)

                client_transport.write(compressed)

                waiter = self._loop.create_future()
                self._waiters[event_name].append(waiter)

                await waiter


            while bool(self.queue[event_name]) and self._stream:

                (
                    _,
                    shard_id,
                    _,
                    response_data,
                    _, 
                    _
                ) = self.queue[event_name].pop()
            
                yield(
                    shard_id,
                    response_data
                )


        self.queue.clear()

    def read(
        self,
        data: bytes,
        transport: asyncio.Transport
    ) -> None:
        decompressed = b''

        try:
            decompressed = self._decompressor.decompress(data)

        except Exception as decompression_error:
            self._pending_responses.append(
                asyncio.create_task(
                    self._send_error(
                        error_message=str(decompression_error),
                        transport=transport
                    )
                )
            )

            if bool(self._last_call):
                event_name = self._last_call.pop()
                event_waiter = self._waiters[event_name]

                if bool(event_waiter):
                    waiter = event_waiter.pop()

                    try:

                        waiter.set_result(None)

                    except asyncio.InvalidStateError:
                        pass

            return

        decrypted = self._encryptor.decrypt(decompressed)

        result: Tuple[
            str, 
            int, 
            float, 
            Any, 
            str, 
            int
        ] = pickle.loads(decrypted)

        (
            message_type, 
            shard_id, 
            event_name,
            payload, 
            incoming_host, 
            incoming_port
        ) = result

        if message_type == 'request':
            self._pending_responses.append(
                asyncio.create_task(
                    self._read(
                        event_name,
                        self.events.get(event_name)(
                            shard_id,
                            self.parsers[event_name](**payload)
                        ),
                        transport
                    )
                )
            )

        elif message_type == "stream_connect":

            self.queue[event_name].append((
                message_type, 
                shard_id,
                event_name,
                payload, 
                incoming_host,
                incoming_port
            ))

            self._pending_responses.append(
                asyncio.create_task(
                    self._initialize_stream(
                        event_name,
                        transport
                    )
                )
            )

            event_waiter = self._waiters[event_name]

            if bool(event_waiter):
                waiter = event_waiter.pop()

                try:

                    waiter.set_result(None)

                except asyncio.InvalidStateError:
                    pass

        elif message_type == 'stream' or message_type == "stream_connect":

            self.queue[event_name].append((
                message_type, 
                shard_id,
                event_name,
                payload, 
                incoming_host,
                incoming_port
            ))

            self._pending_responses.append(
                asyncio.create_task(
                    self._read_iterator(
                        event_name,
                        self.events.get(event_name)(
                            shard_id,
                            self.parsers[event_name](**payload)
                        ),
                        transport
                    )
                )
            )

            event_waiter = self._waiters[event_name]

            if bool(event_waiter):
                waiter = event_waiter.pop()

                try:

                    waiter.set_result(None)

                except asyncio.InvalidStateError:
                    pass

        else:

            if event_name is None and bool(self._last_call):
                event_name = self._last_call.pop()

                
            event_waiter = self._waiters[event_name]

            if bool(event_waiter):
                waiter = event_waiter.pop()

                try:

                    waiter.set_result((
                        message_type, 
                        shard_id,
                        event_name,
                        payload, 
                        incoming_host,
                        incoming_port
                    ))

                except asyncio.InvalidStateError:
                    pass

    async def _read(
        self,
        event_name: str,
        coroutine: Coroutine,
        transport: asyncio.Transport
    ) -> Coroutine[Any, Any, None]:
        response: Message = await coroutine

        item = pickle.dumps(
            (
                'response', 
                self.id_generator.generate(),
                event_name,
                response.to_data(), 
                self.host,
                self.port
            ),
            protocol=pickle.HIGHEST_PROTOCOL
        )

        encrypted_message = self._encryptor.encrypt(item)
        compressed = self._compressor.compress(encrypted_message)

        transport.write(compressed)

    async def _read_iterator(
        self,
        event_name: str,
        coroutine: AsyncIterable[Message],
        transport: asyncio.Transport       
    ) -> Coroutine[Any, Any, None]:
  
        async for response in coroutine:
            item = pickle.dumps(
                (
                    'response', 
                    self.id_generator.generate(),
                    event_name,
                    response.to_data(), 
                    self.host,
                    self.port
                ),
                protocol=pickle.HIGHEST_PROTOCOL
            )

            encrypted_message = self._encryptor.encrypt(item)
            compressed = self._compressor.compress(encrypted_message)

            transport.write(compressed)

    async def _initialize_stream(
        self,
        event_name: str,
        transport: asyncio.Transport            
    ) -> Coroutine[Any, Any, None]:
        message = Message()
        item = pickle.dumps(
            (
                'response', 
                self.id_generator.generate(),
                event_name,
                message.to_data(), 
                self.host,
                self.port
            ),
            protocol=pickle.HIGHEST_PROTOCOL
        )

        encrypted_message = self._encryptor.encrypt(item)
        compressed = self._compressor.compress(encrypted_message)

        transport.write(compressed)

    async def _send_error(
        self,
        error_message: str,
        transport: asyncio.Transport
    ) -> Coroutine[Any, Any, None]:
        
        error = Message(
            error=error_message
        )

        item = pickle.dumps(
            (
                'response', 
                self.id_generator.generate(),
                None,
                error.to_data(), 
                self.host,
                self.port
            ),
            protocol=pickle.HIGHEST_PROTOCOL
        )

        encrypted_message = self._encryptor.encrypt(item)
        compressed = self._compressor.compress(encrypted_message)

        transport.write(compressed)
        
    async def close(self) -> None:
        self._stream = False
        self._running = False

        for client in self._client_transports.values():
            client.abort()
        
        if self._cleanup_task:
            self._cleanup_task.cancel()
            if self._cleanup_task.cancelled() is False:
                try:
                    self._sleep_task.cancel()
                    if not self._sleep_task.cancelled():
                        await self._sleep_task

                except asyncio.CancelledError:
                    pass

                try:

                    await self._cleanup_task

                except asyncio.CancelledError:
                    pass

            